@@ -3,7 +3,7 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2009-2022, VU University Amsterdam
+    Copyright (c)  2009-2023, VU University Amsterdam
 			      SWI-Prolog Solutions b.v.
     All rights reserved.
 
@@ -79,7 +79,7 @@ syntax_error(const char *culprit)
        PL_unify_term(ex,
 		     PL_FUNCTOR, FUNCTOR_error2,
 		       PL_FUNCTOR, FUNCTOR_syntax_error1,
-		         PL_CHARS, culprit,
+			 PL_CHARS, culprit,
 		       PL_VARIABLE) )
     return PL_raise_exception(ex);
 
@@ -95,7 +95,7 @@ type_error(const char *expected, term_t found)
        PL_unify_term(ex,
 		     PL_FUNCTOR, FUNCTOR_error2,
 		       PL_FUNCTOR, FUNCTOR_type_error2,
-		         PL_CHARS, expected,
+			 PL_CHARS, expected,
 			 PL_TERM, found,
 		       PL_VARIABLE) )
     return PL_raise_exception(ex);
@@ -112,7 +112,7 @@ domain_error(const char *expected, term_t found)
        PL_unify_term(ex,
 		     PL_FUNCTOR, FUNCTOR_error2,
 		       PL_FUNCTOR, FUNCTOR_domain_error2,
-		         PL_CHARS, expected,
+			 PL_CHARS, expected,
 			 PL_TERM, found,
 		       PL_VARIABLE) )
     return PL_raise_exception(ex);
@@ -507,7 +507,13 @@ add_normalized_range_charbuf(charbuf *cb, const range *r, int iri, int flags)
     { const pl_wchar_t *e;
 
       if ( (e=get_encoded_utf8(s, &c)) )
-      { s = e;
+      { if ( flags == ESC_QUERY &&
+	     (c == '=' || c == '&' || c == ';') )
+	{ while( s<e )
+	    add_charbuf(cb, *s++);
+	  continue;
+	}
+	s = e;
       } else if (hex(s+1, 2, &c) )
       { s += 3;
       } else
@@ -1051,12 +1057,31 @@ unify_uri_authority_components(term_t components,
     }
   }
   host.start = s;
+  if ( *s == '[' )
+  { host.end = skip_not(s+1, end, L"]");
+    if ( host.end == end )	/* No "]" */
+      goto no_ipv6;
+    if ( host.end == end-1 )	/* ends with "]" */
+    { host.start++;
+      goto done;
+    }
+    if ( host.end+1 < end && host.end[1] == ':' )
+    { host.start++;
+      port.start = host.end+2;
+      port.end = end;
+      goto done;
+    }
+    s = host.end;
+  }
+
+no_ipv6:
   host.end = skip_not(s, end, L":");
   if ( host.end < end )
   { port.start = host.end+1;
     port.end = end;
   }
 
+done:
   if ( user.start )
     unify_decoded_atom(av+0, &user, ESC_USER);
   if ( passwd.start )
@@ -1095,6 +1120,7 @@ uri_authority_components(term_t Authority, term_t components)
     int rc;
 
     init_charbuf(&b);
+    /* user[:password] */
     if ( (rc=get_text_arg(components, 1, &len, &s, TXT_EX_TEXT)) == TRUE )
     { add_nchars_charbuf(&b, len, s);
       if ( (rc=get_text_arg(components, 2, &len, &s, TXT_EX_TEXT)) == TRUE )
@@ -1109,12 +1135,20 @@ uri_authority_components(term_t Authority, term_t components)
     { free_charbuf(&b);
       return FALSE;
     }
+    /* host */
     if ( (rc=get_text_arg(components, 3, &len, &s, TXT_EX_TEXT)) == TRUE )
-    { add_nchars_charbuf(&b, len, s);
+    { if ( wcschr(s, ':') )
+      { add_charbuf(&b, '[');
+	add_nchars_charbuf(&b, len, s);
+	add_charbuf(&b, ']');
+      } else
+      { add_nchars_charbuf(&b, len, s);
+      }
     } else if ( rc == -1 )
     { free_charbuf(&b);
       return FALSE;
     }
+    /* port */
     if ( (rc=get_text_arg(components, 4, &len, &s,
 			  TXT_EX_TEXT|CVT_INTEGER)) == TRUE )
     { add_charbuf(&b, ':');
