@@ -3,9 +3,10 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2013-2018, University of Amsterdam
+    Copyright (c)  2013-2022, University of Amsterdam
                               VU University Amsterdam
                               CWI, Amsterdam
+                              SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -54,7 +55,8 @@
 :- use_module(library(main)).
 :- use_module(library(readutil)).
 
-:- if(exists_source(library(http/http_ssl_plugin))).
+:- if(( exists_source(library(http/http_ssl_plugin)),
+        \+ current_prolog_flag(pldoc_to_tex,true))).
 :- use_module(library(ssl)).
 :- use_module(library(http/http_ssl_plugin)).
 :- endif.
@@ -515,13 +517,16 @@ options_password(Options, Passwd) :-
     split_string(String, "", "\r\n\t ", [Passwd]).
 options_password(_, '').
 
-%!  start_server(+Server) is det.
+%!  start_servers(+Servers) is det.
 %
 %   Start the HTTP server.  It performs the following steps:
 %
 %     1. Call broadcast(http(pre_server_start))
-%     2. Call http_server(http_dispatch, Options)
-%     2. Call broadcast(http(post_server_start))
+%     2. Foreach server
+%        a. Call broadcast(http(pre_server_start(Port)))
+%        b. Call http_server(http_dispatch, Options)
+%        c. Call broadcast(http(post_server_start(Port)))
+%     3. Call broadcast(http(post_server_start))
 %
 %   This predicate can be  hooked   using  http_server_hook/1.  This
 %   predicate is executed after
@@ -544,7 +549,10 @@ start_server(server(_Scheme, Socket, Options)) :-
     http_server_hook([tcp_socket(Socket)|Options]),
     !.
 start_server(server(_Scheme, Socket, Options)) :-
-    http_server(http_dispatch, [tcp_socket(Socket)|Options]).
+    option(port(Port), Options),
+    broadcast(http(pre_server_start(Port))),
+    http_server(http_dispatch, [tcp_socket(Socket)|Options]),
+    broadcast(http(post_server_start(Port))).
 
 make_socket(server(Scheme, Address, Options),
             server(Scheme, Socket, Options)) :-
@@ -794,7 +802,7 @@ setup_signals(Options) :-
 
 quit(Signal) :-
     debug(daemon, 'Dying on signal ~w', [Signal]),
-    thread_send_message(main, quit).
+    thread_send_message(main, quit(Signal)).
 
 reload(Signal) :-
     debug(daemon, 'Reload on signal ~w', [Signal]),
@@ -827,11 +835,14 @@ wait(Options) :-
     repeat,
         State = deadline(Count),
         Deadline is FirstDeadline+Count*Interval,
-        (   thread_get_message(Me, Msg, [deadline(Deadline)])
+        (   thread_idle(thread_get_message(Me, Msg, [deadline(Deadline)]),
+                        long)
         ->  catch(ignore(handle_message(Msg)), E,
                   print_message(error, E)),
-            Msg == quit,
-            halt(0)
+            Msg = quit(Signal),
+            catch(broadcast(http(shutdown)), E,
+                  print_message(error, E)),
+            halt(Signal)
         ;   Count1 is Count + 1,
             nb_setarg(1, State, Count1),
             catch(broadcast(maintenance(Interval, Deadline)), E,
@@ -841,7 +852,7 @@ wait(Options) :-
 wait(_) :-
     thread_self(Me),
     repeat,
-        thread_get_message(Me, Msg),
+        thread_idle(thread_get_message(Me, Msg), long),
         catch(ignore(handle_message(Msg)), E,
               print_message(error, E)),
         Msg == quit,
