@@ -60,13 +60,9 @@ please visit https://www.swi-prolog.org/pldoc/package/plunit.
 :- autoload(library(apply),
             [maplist/3, include/3, maplist/2, foldl/4, partition/4]).
 :- autoload(library(lists), [member/2, append/2, flatten/2, append/3]).
-:- autoload(library(option),
-            [ option/3, option/2, merge_options/3, select_option/4,
-              select_option/3
-            ]).
+:- autoload(library(option), [ option/3, option/2, select_option/3 ]).
 :- autoload(library(ordsets), [ord_intersection/3]).
 :- autoload(library(error), [must_be/2, domain_error/2]).
-:- autoload(library(thread), [concurrent_forall/2]).
 :- autoload(library(aggregate), [aggregate_all/3]).
 :- autoload(library(streams), [with_output_to/3]).
 :- autoload(library(time), [call_with_time_limit/2]).
@@ -84,17 +80,40 @@ swi     :- catch(current_prolog_flag(dialect, swi), _, fail), !.
 swi     :- catch(current_prolog_flag(dialect, yap), _, fail).
 sicstus :- catch(current_prolog_flag(system_type, _), _, fail).
 
-
-:- if(swi).
 throw_error(Error_term,Impldef) :-
     throw(error(Error_term,context(Impldef,_))).
 
 :- set_prolog_flag(generate_debug_info, false).
-current_test_flag(Name, Value) :-
-    current_prolog_flag(Name, Value).
+current_test_flag(optimise, Value) =>
+    current_prolog_flag(optimise, Value).
+current_test_flag(occurs_check, Value) =>
+    (   current_prolog_flag(plunit_occurs_check, Value0)
+    ->  Value = Value0
+    ;   current_prolog_flag(occurs_check, Value)
+    ).
+current_test_flag(Name, Value), atom(Name) =>
+    atom_concat(plunit_, Name, Flag),
+    current_prolog_flag(Flag, Value).
+current_test_flag(Name, Value), var(Name) =>
+    global_test_option(Opt, _, _Type, _Default),
+    functor(Opt, Name, 1),
+    current_test_flag(Name, Value).
 
 set_test_flag(Name, Value) :-
-    create_prolog_flag(Name, Value, []).
+    Opt =.. [Name, Value],
+    global_test_option(Opt),
+    !,
+    atom_concat(plunit_, Name, Flag),
+    set_prolog_flag(Flag, Value).
+set_test_flag(Name, _) :-
+    domain_error(test_flag, Name).
+
+current_test_flags(Flags) :-
+    findall(Flag, current_test_flag(Flag), Flags).
+
+current_test_flag(Opt) :-
+    current_test_flag(Name, Value),
+    Opt =.. [Name, Value].
 
 % ensure expansion to avoid tracing
 goal_expansion(forall(C,A),
@@ -102,71 +121,22 @@ goal_expansion(forall(C,A),
 goal_expansion(current_module(Module,File),
 	       module_property(Module, file(File))).
 
-:- if(current_prolog_flag(dialect, yap)).
-
-'$set_predicate_attribute'(_, _, _).
-
-:- endif.
-:- endif.
-
-:- if(sicstus).
-throw_error(Error_term,Impldef) :-
-    throw(error(Error_term,i(Impldef))). % SICStus 3 work around
-
-% SWI-Compatibility
-:- op(700, xfx, =@=).
-
-'$set_source_module'(_, _).
-
-%!  current_test_flag(?Name, ?Value) is nondet.
-%
-%   Query  flags  that  control  the    testing   process.  Emulates
-%   SWI-Prologs flags.
-
-:- dynamic test_flag/2. % Name, Val
-
-current_test_flag(optimise, Val) :-
-    current_prolog_flag(compiling, Compiling),
-    (   Compiling == debugcode ; true % TBD: Proper test
-    ->  Val = false
-    ;   Val = true
-    ).
-current_test_flag(Name, Val) :-
-    test_flag(Name, Val).
-
-
-%!  set_test_flag(+Name, +Value) is det.
-
-set_test_flag(Name, Val) :-
-    var(Name),
-    !,
-    throw_error(instantiation_error, set_test_flag(Name,Val)).
-set_test_flag( Name, Val ) :-
-    retractall(test_flag(Name,_)),
-    asserta(test_flag(Name, Val)).
-
-:- op(1150, fx, thread_local).
-
-:- discontiguous
-    user:term_expansion/2.
-
-user:term_expansion((:- thread_local(PI)), (:- dynamic(PI))) :-
-    prolog_load_context(module, plunit).
-
-:- endif.
 
 		 /*******************************
 		 *            IMPORTS           *
 		 *******************************/
 
-:- initialization
-   (   current_test_flag(test_options, _)
-   ->  true
-   ;   set_test_flag(test_options,
-                     [ run(make),       % run tests on make/0
-                       output(on_failure)
-                     ])
-   ).
+:- initialization init_flags.
+
+init_flags :-
+    (   global_test_option(Option, _Value, _Type, Default),
+	Default \== (-),
+	Option =.. [Name,_],
+	atom_concat(plunit_, Name, Flag),
+	create_prolog_flag(Flag, Default, [keep(true)]),
+	fail
+    ;   true
+    ).
 
 %!  set_test_options(+Options)
 %
@@ -194,6 +164,9 @@ user:term_expansion((:- thread_local(PI)), (:- dynamic(PI))) :-
 %      suppress all output and if `on_failure`, emit the output
 %      if the test fails.
 %
+%    - show_blocked(+Bool)
+%      Show individual blocked tests during the report.
+%
 %    - occurs_check(+Mode)
 %      Defines the default for the `occurs_check` flag during
 %      testing.
@@ -203,65 +176,56 @@ user:term_expansion((:- thread_local(PI)), (:- dynamic(PI))) :-
 %      of run_tests/1.  Used to improve cooperation with
 %      memory debuggers such as dmalloc.
 %
-%    - concurrent(+Bool)
-%      If `true` (default `false`), run all tests in a unit
-%      concurrently.
-%
 %    - jobs(Num)
-%      Number of jobs to use for concurrent testing.  Defaults
-%      to the number of cores.
+%      Number of jobs to use for concurrent testing.  Default
+%      is one, implying sequential testing.
 %
 %    - timeout(+Seconds)
 %      Set timeout for each individual test.  This acts as a
 %      default that may be overuled at the level of units or
-%      individual tests.
+%      individual tests.   A timeout of 0 or negative is handled
+%      as _inifinite_.
 
 set_test_options(Options) :-
-    select_option(sto(Mode), Options, Options1, _),
-    (   Mode == true
-    ->  print_message(warning, plunit(sto(true)))
-    ;   true
-    ),
-    (   option(jobs(Jobs), Options1),
-	Jobs > 1
-    ->  merge_options([concurrent(true)], Options1, Options2)
-    ;   Options2 = Options1
-    ),
-    valid_options(global_test_option, Options2),
-    (   current_test_flag(test_options, OldOptions)
-    ->  merge_options(Options2, OldOptions, NewOptions),
-        set_test_flag(test_options, NewOptions)
-    ;   set_test_flag(test_options, Options)
-    ).
+    flatten([Options], List),
+    maplist(set_test_option, List).
 
-global_test_option(load(Load)) :-
-    must_be(oneof([never,always,normal]), Load).
-global_test_option(output(Cond)) :-
-    must_be(oneof([always,on_failure]), Cond).
-global_test_option(format(Feedback)) :-
-    must_be(oneof([tty,log]), Feedback).
-global_test_option(run(When)) :-
-    must_be(oneof([manual,make,make(all)]), When).
-global_test_option(silent(Bool)) :-
-    must_be(boolean, Bool).
-global_test_option(occurs_check(Mode)) :-
-    must_be(oneof([false,true,error]), Mode).
-global_test_option(cleanup(Bool)) :-
-    must_be(boolean, Bool).
-global_test_option(concurrent(Bool)) :-
-    must_be(boolean, Bool).
-global_test_option(jobs(Count)) :-
-    must_be(positive_integer, Count).
-global_test_option(timeout(Number)) :-
-    must_be(number, Number).
+set_test_option(sto(true)) =>
+    print_message(warning, plunit(sto(true))).
+set_test_option(jobs(Jobs)) =>
+    must_be(positive_integer, Jobs),
+    set_test_option_flag(jobs(Jobs)).
+set_test_option(Option),
+  compound(Option), global_test_option(Option) =>
+    set_test_option_flag(Option).
+set_test_option(Option) =>
+    domain_error(option, Option).
+
+global_test_option(Opt) :-
+    global_test_option(Opt, Value, Type, _Default),
+    must_be(Type, Value).
+
+global_test_option(load(Load), Load, oneof([never,always,normal]), normal).
+global_test_option(output(Cond), Cond, oneof([always,on_failure]), on_failure).
+global_test_option(format(Feedback), Feedback, oneof([tty,log]), tty).
+global_test_option(silent(Silent), Silent, boolean, false).
+global_test_option(show_blocked(Blocked), Blocked, boolean, false).
+global_test_option(run(When), When, oneof([manual,make,make(all)]), make).
+global_test_option(occurs_check(Mode), Mode, oneof([false,true,error]), -).
+global_test_option(cleanup(Bool), Bool, boolean, true).
+global_test_option(jobs(Count), Count, positive_integer, 1).
+global_test_option(timeout(Number), Number, number, 3600).
+
+set_test_option_flag(Option) :-
+    Option =.. [Name, Value],
+    set_test_flag(Name, Value).
 
 %!  loading_tests
 %
 %   True if tests must be loaded.
 
 loading_tests :-
-    current_test_flag(test_options, Options),
-    option(load(Load), Options, normal),
+    current_test_flag(load, Load),
     (   Load == always
     ->  true
     ;   Load == normal,
@@ -591,7 +555,8 @@ test_set_option(cleanup(X)) :-
 test_set_option(occurs_check(V)) :-
     must_be(oneof([false,true,error]), V).
 test_set_option(concurrent(V)) :-
-    must_be(boolean, V).
+    must_be(boolean, V),
+    print_message(informational, plunit(concurrent)).
 test_set_option(timeout(Seconds)) :-
     must_be(number, Seconds).
 
@@ -602,12 +567,14 @@ test_set_option(timeout(Seconds)) :-
 :- meta_predicate
        reify_tmo(0, -, +),
        reify(0, -),
+       capture_output(0,-),
        capture_output(0,-,+).
 
 %!  reify_tmo(:Goal, -Result, +Options) is det.
 
 reify_tmo(Goal, Result, Options) :-
     option(timeout(Time), Options),
+    Time > 0,
     !,
     reify(call_with_time_limit(Time, Goal), Result0),
     (   Result0 = throw(time_limit_exceeded)
@@ -631,6 +598,10 @@ reify(Goal, Result) :-
     ;   Result = false
     ).
 
+capture_output(Goal, Output) :-
+    current_test_flag(output, OutputMode),
+    capture_output(Goal, Output, [output(OutputMode)]).
+
 capture_output(Goal, Output, Options) :-
     option(output(How), Options, always),
     (   How == always
@@ -647,6 +618,7 @@ capture_output(Goal, Output, Options) :-
 		 *******************************/
 
 :- dynamic
+    output_streams/2,               % Output, Error
     test_count/1,                   % Count
     passed/5,                       % Unit, Test, Line, Det, Time
     failed/5,                       % Unit, Test, Line, Reason, Time
@@ -699,49 +671,52 @@ run_tests(all, Options) :-
     run_tests(Units, Options).
 run_tests(Set, Options) :-
     valid_options(global_test_option, Options, Global, Rest),
-    current_test_flag(test_options, Old),
+    current_test_flags(Old),
     setup_call_cleanup(
 	set_test_options(Global),
 	( flatten([Set], List),
 	  maplist(runnable_tests, List, Units),
 	  with_mutex(plunit, run_tests_sync(Units, Rest))
 	),
-	set_test_flag(test_options, Old)).
+	set_test_options(Old)).
 
-run_tests_sync(Units, Options) :-
+run_tests_sync(Units0, Options) :-
     cleanup,
-    count_tests(Units, Count),
+    count_tests(Units0, Units, Count),
     asserta(test_count(Count)),
+    save_output_state,
     setup_call_cleanup(
 	setup_jobs(Count),
 	setup_call_cleanup(
 	    setup_trap_assertions(Ref),
-	    run_units_and_check_errors(Units, Options),
-	    report_and_cleanup(Ref, Options)),
-	cleanup_jobs).
+	    ( call_time(run_units(Units, Options), Time),
+              test_summary(_All, Summary)
+            ),
+	    report_and_cleanup(Ref, Time, Options)),
+	cleanup_jobs),
+    (   option(summary(Summary), Options)
+    ->  true
+    ;   test_summary_passed(Summary) % fail if some test failed
+    ).
 
-%!  report_and_cleanup(+Ref, +Options)
+%!  report_and_cleanup(+Ref, +Time, +Options)
 %
 %   Undo changes to the environment   (trapping  assertions), report the
 %   results and cleanup.
 
-report_and_cleanup(Ref, _Options) :-
+report_and_cleanup(Ref, Time, Options) :-
     cleanup_trap_assertions(Ref),
-    report,
+    report(Time, Options),
     cleanup_after_test.
 
+
 %!  run_units_and_check_errors(+Units, +Options) is semidet.
 %
 %   Run all test units and succeed if all tests passed.
 
-run_units_and_check_errors(Units, Options) :-
+run_units(Units, _Options) :-
     maplist(schedule_unit, Units),
-    job_wait,
-    test_summary(_All, Summary),
-    (   option(summary(Summary), Options)
-    ->  true
-    ;   test_summary_passed(Summary)
-    ).
+    job_wait(_).
 
 %!  runnable_tests(+Spec, -Plan) is det.
 %
@@ -770,12 +745,12 @@ runnable_tests(Spec, Unit:RunnableTests) :-
                 RunnableTests)
     ).
 
-runnable_test(Unit, Test, Module, @(Test,Line)) :-
-    current_test(Unit, Test, Line, _Body, TestOptions),
+runnable_test(Unit, Name, Module, @(Test,Line)) :-
+    current_test(Unit, Name, Line, _Body, TestOptions),
     (   option(blocked(Reason), TestOptions)
-    ->  assert(blocked(Unit, Test, Line, Reason)),
-        fail
-    ;   condition(Module, test(Unit,Test,Line), TestOptions)
+    ->  Test = blocked(Name, Reason)
+    ;   condition(Module, test(Unit,Name,Line), TestOptions),
+        Test = Name
     ).
 
 unit_from_spec(Unit0:Tests0, Unit, Tests, Module, Options), atom(Unit0) =>
@@ -792,18 +767,36 @@ unit_from_spec(Unit0, Unit, _, Module, Options), atom(Unit0) =>
     ;   throw_error(existence_error(unit_test, Unit), _)
     ).
 
-%!  count_tests(+Units, -Count) is det.
+%!  count_tests(+Units0, -Units, -Count) is det.
 %
 %   Count the number of tests to   run. A forall(Generator, Test) counts
 %   as a single test. During the execution,   the  concrete tests of the
 %   _forall_ are considered "sub tests".
 
-count_tests(Units, Count) :-
-    foldl(count_tests_in_unit, Units, 0, Count).
+count_tests(Units0, Units, Count) :-
+    count_tests(Units0, Units, 0, Count).
 
-count_tests_in_unit(_Unit:Tests, Count0, Count) :-
-    length(Tests, N),
-    Count is Count0+N.
+count_tests([], T, C0, C) =>
+    T = [],
+    C = C0.
+count_tests([_:[]|T0], T, C0, C) =>
+    count_tests(T0, T, C0, C).
+count_tests([Unit:Tests|T0], T, C0, C) =>
+    partition(is_blocked, Tests, Blocked, Use),
+    maplist(assert_blocked(Unit), Blocked),
+    (   Use == []
+    ->  count_tests(T0, T, C0, C)
+    ;   length(Use, N),
+        C1 is C0+N,
+        T = [Unit:Use|T1],
+        count_tests(T0, T1, C1, C)
+    ).
+
+is_blocked(@(blocked(_,_),_)) => true.
+is_blocked(_) => fail.
+
+assert_blocked(Unit, @(blocked(Test, Reason), Line)) =>
+    assert(blocked(Unit, Test, Line, Reason)).
 
 %!  run_unit(+Unit) is det.
 %
@@ -832,16 +825,6 @@ end_unit(Unit, Summary) :-
     job_info(end(unit(Unit, Summary))),
     job_feedback(informational, end(Unit, Summary)).
 
-:- if(current_prolog_flag(threads, true)).
-run_unit_2(Unit, Tests) :-
-    unit_options(Unit, UnitOptions),
-    option(concurrent(true), UnitOptions, false),
-    current_test_flag(test_options, GlobalOptions),
-    option(concurrent(true), GlobalOptions),
-    !,
-    concurrent_forall(member(Test, Tests),
-                      run_test(Unit, Test)).
-:- endif.
 run_unit_2(Unit, Tests) :-
     forall(member(Test, Tests),
 	   run_test(Unit, Test)).
@@ -853,6 +836,7 @@ unit_options(Unit, Options) :-
 
 cleanup :-
     set_flag(plunit_test, 1),
+    retractall(output_streams(_,_)),
     retractall(test_count(_)),
     retractall(passed(_, _, _, _, _)),
     retractall(failed(_, _, _, _, _)),
@@ -864,9 +848,7 @@ cleanup :-
     retractall(forall_failures(_,_)).
 
 cleanup_after_test :-
-    current_test_flag(test_options, Options),
-    option(cleanup(Cleanup), Options, false),
-    (   Cleanup == true
+    (   current_test_flag(cleanup, true)
     ->  cleanup
     ;   true
     ).
@@ -904,8 +886,7 @@ unit_in_files(Files, Unit) :-
 %   Called indirectly from make/0 after Files have been reloaded.
 
 make_run_tests(Files) :-
-    current_test_flag(test_options, Options),
-    option(run(When), Options, manual),
+    current_test_flag(run, When),
     (   When == make
     ->  run_tests_in_files(Files)
     ;   When == make(all)
@@ -937,9 +918,8 @@ test_assertion_failed(Reason, Goal) :-
     ->  true
     ;   AssertLoc = unknown
     ),
-    current_test_flag(test_options, Options),
     report_failed_assertion(Unit:Test, Line, AssertLoc,
-			    Progress, Reason, Goal, Options),
+			    Progress, Reason, Goal),
     assert_cyclic(failed_assertion(Unit, Test, Line, AssertLoc,
 				   Progress, Reason, Goal)).
 
@@ -951,7 +931,7 @@ assertion_location(Stack, File:Line) :-
     prolog_stack_frame_property(CallerFrame, location(File:Line)).
 
 report_failed_assertion(UnitTest, Line, AssertLoc,
-			Progress, Reason, Goal, _Options) :-
+			Progress, Reason, Goal) :-
     print_message(
 	error,
 	plunit(failed_assertion(UnitTest, Line, AssertLoc,
@@ -990,7 +970,7 @@ run_test(Unit, Name, Line, UnitOptions, Options, Body) :-
     term_variables(Generator, Vars),
     start_test(Unit, @(Name,Line), Nth),
     State = state(0),
-    call_time(forall(Module:Generator,                      % may become concurrent
+    call_time(forall(Module:Generator,            % may become concurrent
                      (   incr_forall(State, I),
                          run_test_once6(Unit, Name, forall(Vars, Nth-I), Line,
                                         UnitOptions, Options, Body)
@@ -1016,19 +996,23 @@ incr_forall(State, I) :-
 %   Inherit the `timeout` and `occurs_check` option (Global -> Unit -> Test).
 
 run_test_once6(Unit, Name, Progress, Line, UnitOptions, Options, Body) :-
-    current_test_flag(test_options, GlobalOptions),
-    inherit_option(timeout,      Options,  [UnitOptions, GlobalOptions], Options1),
-    inherit_option(occurs_check, Options1, [UnitOptions, GlobalOptions], Options2),
+    current_test_flag(timeout, DefTimeOut),
+    current_test_flag(occurs_check, DefOccurs),
+    inherit_option(timeout,      Options,  [UnitOptions], DefTimeOut, Options1),
+    inherit_option(occurs_check, Options1, [UnitOptions], DefOccurs, Options2),
     run_test_once(Unit, Name, Progress, Line, Options2, Body).
 
-inherit_option(Name, Options0, Chain, Options) :-
+inherit_option(Name, Options0, Chain, Default, Options) :-
     Term =.. [Name,_Value],
     (   option(Term, Options0)
     ->  Options = Options0
     ;   member(Opts, Chain),
         option(Term, Opts)
     ->  Options = [Term|Options0]
-    ;   Options = Options0
+    ;   Default == (-)
+    ->  Options = Options0
+    ;   Opt =.. [Name,Default],
+	Options = [Opt|Options0]
     ).
 
 %!  run_test_once(+Unit, +Name, Progress, +Line, +Options, +Body)
@@ -1037,23 +1021,21 @@ inherit_option(Name, Options0, Chain, Options) :-
 %   unification settings wrt. the occurs check.
 
 run_test_once(Unit, Name, Progress, Line, Options, Body) :-
-    option(occurs_check(Mode), Options),
+    option(occurs_check(Occurs), Options),
     !,
-    current_test_flag(test_options, GlobalOptions),
     begin_test(Unit, Name, Line, Progress),
     current_prolog_flag(occurs_check, Old),
     setup_call_cleanup(
-	set_prolog_flag(occurs_check, Mode),
+	set_prolog_flag(occurs_check, Occurs),
 	capture_output(run_test_6(Unit, Name, Line, Options, Body, Result),
-		       Output, GlobalOptions),
+		       Output),
 	set_prolog_flag(occurs_check, Old)),
     end_test(Unit, Name, Line, Progress),
     report_result(Result, Progress, Output, Options).
 run_test_once(Unit, Name, Progress, Line, Options, Body) :-
-    current_test_flag(test_options, GlobalOptions),
     begin_test(Unit, Name, Line, Progress),
     capture_output(run_test_6(Unit, Name, Line, Options, Body, Result),
-		   Output, GlobalOptions),
+		   Output),
     end_test(Unit, Name, Line, Progress),
     report_result(Result, Progress, Output, Options).
 
@@ -1284,9 +1266,7 @@ match_error(Expect, Rec) :-
 setup(Module, Context, Options) :-
     option(setup(Setup), Options),
     !,
-    current_test_flag(test_options, GlobalOptions),
-    capture_output(reify(call_ex(Module, Setup), Result),
-		   Output, GlobalOptions),
+    capture_output(reify(call_ex(Module, Setup), Result), Output),
     (   Result == true
     ->  true
     ;   print_message(error,
@@ -1302,9 +1282,7 @@ setup(_,_,_).
 condition(Module, Context, Options) :-
     option(condition(Cond), Options),
     !,
-    current_test_flag(test_options, GlobalOptions),
-    capture_output(reify(call_ex(Module, Cond), Result),
-		   Output, GlobalOptions),
+    capture_output(reify(call_ex(Module, Cond), Result), Output),
     (   Result == true
     ->  true
     ;   Result == false
@@ -1353,7 +1331,7 @@ success(Unit, Name, Progress, Line, Det, Time, _Output, Options) :-
     flush_output(user_error),
     assert(fixme(Unit, Name, Line, Reason, Ok)).
 success(Unit, Name, Progress, Line, _, Time, Output, Options) :-
-    failed_assertion(Unit, Name, Line, _,_,_,_),
+    failed_assertion(Unit, Name, Line, _,Progress,_,_),
     !,
     failure(Unit, Name, Progress, Line, assertion, Time, Output, Options).
 success(Unit, Name, Progress, Line, Det, Time, _Output, Options) :-
@@ -1439,10 +1417,11 @@ schedule_unit(Unit) :-
 %   Setup threads for concurrent testing.
 
 setup_jobs(Count) :-
-    current_prolog_flag(cpu_count, Cores),
-    current_test_flag(test_options, Options),
-    option(concurrent(true), Options),
-    option(jobs(Jobs0), Options, Cores),
+    (   current_test_flag(jobs, Jobs0),
+	integer(Jobs0)
+    ->  true
+    ;   current_prolog_flag(cpu_count, Jobs0)
+    ),
     Jobs is min(Count, Jobs0),
     Jobs > 1,
     !,
@@ -1472,6 +1451,8 @@ plunit_job(Queue) :-
 
 job(unit(Unit:Tests)) =>
     run_unit(Unit:Tests).
+job(test(Unit, Test)) =>
+    run_test(Unit, Test).
 
 cleanup_jobs :-
     retract(job_data(Queue, TIDSs)),
@@ -1480,24 +1461,23 @@ cleanup_jobs :-
     maplist(thread_join, TIDSs).
 cleanup_jobs.
 
-%!  job_wait is det.
+%!  job_wait(?Unit) is det.
 %
 %   Wait for all test jobs to finish.
 
-job_wait :-
-    thread_wait(\+ scheduled_unit(_),
+job_wait(Unit) :-
+    thread_wait(\+ scheduled_unit(Unit),
 		[ wait_preds([scheduled_unit/1]),
 		  timeout(1)
 		]),
     !.
-job_wait :-
+job_wait(Unit) :-
     job_data(_Queue, TIDs),
     member(TID, TIDs),
     thread_property(TID, status(running)),
     !,
-    job_wait.
-job_wait.
-
+    job_wait(Unit).
+job_wait(_).
 
 
 job_info(begin(unit(_Unit))) =>
@@ -1513,7 +1493,7 @@ schedule_unit(Unit) :-
 setup_jobs(_) :-
     print_message(silent, plunit(jobs(1))).
 cleanup_jobs.
-job_wait.
+job_wait(_).
 job_info(_).
 
 :- endif.
@@ -1594,51 +1574,55 @@ test_summary(Unit, Summary) :-
     count(timeout(Unit, _0Test, _0Line, _Limit, _0Time), Timeout),
     count(passed(Unit, _0Test, _0Line, _Det, _0Time), Passed),
     count(blocked(Unit, _0Test, _0Line, _0Reason), Blocked),
+    count(fixme(Unit, _0Test, _0Line, _0Reason, _0How), Fixme),
     test_count(Total),
     Summary = plunit{total:Total,
 		     passed:Passed,
 		     failed:Failed,
 		     timeout:Timeout,
-		     blocked:Blocked}.
+		     blocked:Blocked,
+		     fixme:Fixme}.
 
 test_summary_passed(Summary) :-
     _{failed: 0} :< Summary.
 
-%!  report is det.
+%!  report(+Time, +Options) is det.
 %
 %   Print a summary of the tests that ran.
 
-report :-
+report(Time, _Options) :-
     test_summary(_, Summary),
     print_message(silent, plunit(Summary)),
     _{ passed:Passed,
        failed:Failed,
        timeout:Timeout,
-       blocked:Blocked
+       blocked:Blocked,
+       fixme:Fixme
      } :< Summary,
-    (   Passed+Failed+Timeout+Blocked =:= 0
+    (   Passed+Failed+Timeout+Blocked+Fixme =:= 0
     ->  info(plunit(no_tests))
-    ;   Failed+Timeout+Blocked =:= 0
-    ->  report_fixme,
+    ;   Failed+Timeout =:= 0
+    ->  report_blocked(Blocked),
+	report_fixme,
         test_count(Total),
-	info(plunit(all_passed(Total, Passed)))
+	info(plunit(all_passed(Total, Passed, Time)))
     ;   report_blocked(Blocked),
 	report_fixme,
 	report_failed(Failed),
 	report_timeout(Timeout),
-	info(plunit(passed(Passed)))
+	info(plunit(passed(Passed))),
+        info(plunit(total_time(Time)))
     ).
 
 report_blocked(0) =>
     true.
 report_blocked(Blocked) =>
-    info(plunit(blocked(Blocked))),
-    (   blocked(Unit, Name, Line, Reason),
-	unit_file(Unit, File),
-	print_message(informational,
-		      plunit(blocked(File:Line, Name, Reason))),
-	fail ; true
-    ).
+    findall(blocked(Unit:Name, File:Line, Reason),
+	    ( blocked(Unit, Name, Line, Reason),
+	      unit_file(Unit, File)
+	    ),
+	    BlockedTests),
+    info(plunit(blocked(Blocked, BlockedTests))).
 
 report_failed(Failed) :-
     print_message(error, plunit(failed(Failed))).
@@ -1687,16 +1671,22 @@ test_report(What) :-
 		 *******************************/
 
 %!  unit_file(+Unit, -File) is det.
-%!  unit_file(-Unit, +File) is nondet.
+%!  unit_file(?Unit, ?File) is nondet.
+%
+%   True when the test unit Unit is defined in File.
 
-unit_file(Unit, File) :-
-    current_unit(Unit, Module, _Context, _Options),
-    current_module(Module, File),
+unit_file(Unit, File), nonvar(Unit) =>
+    unit_file_(Unit, File),
     !.
-unit_file(Unit, PlFile) :-
-    nonvar(PlFile),
+unit_file(Unit, File) =>
+    unit_file_(Unit, File).
+
+unit_file_(Unit, File) :-
+    current_unit(Unit, Module, _Context, _Options),
+    module_property(Module, file(File)).
+unit_file_(Unit, PlFile) :-
     test_file_for(TestFile, PlFile),
-    current_module(Module, TestFile),
+    module_property(Module, file(TestFile)),
     current_unit(Unit, Module, _Context, _Options).
 
 
@@ -1706,25 +1696,35 @@ unit_file(Unit, PlFile) :-
 
 %!  load_test_files(+Options) is det.
 %
-%   Load .plt test-files related to loaded source-files.
+%   Load .plt test-files related  to   loaded  source-files.  Options is
+%   currently ignored.
 
 load_test_files(_Options) :-
+    State = state(0,0),
     (   source_file(File),
 	file_name_extension(Base, Old, File),
 	Old \== plt,
 	file_name_extension(Base, plt, TestFile),
 	exists_file(TestFile),
+        inc_arg(1, State),
 	(   test_file_for(TestFile, File)
 	->  true
 	;   load_files(TestFile,
 		       [ if(changed),
 			 imports([])
 		       ]),
+            inc_arg(2, State),
 	    asserta(test_file_for(TestFile, File))
 	),
-	fail ; true
+        fail
+    ;   State = state(Total, Loaded),
+        print_message(informational, plunit(test_files(Total, Loaded)))
     ).
 
+inc_arg(Arg, State) :-
+    arg(Arg, State, N0),
+    N is N0+1,
+    nb_setarg(Arg, State, N).
 
 
 		 /*******************************
@@ -1782,16 +1782,14 @@ update_forall_failures(Nth, _) =>
     asserta(forall_failures(Nth, Failed)).
 
 message_level(Level) :-
-    current_test_flag(test_options, Options),
-    option(silent(Silent), Options, false),
-    (   Silent == false
-    ->  Level = informational
-    ;   Level = silent
+    (   current_test_flag(silent, true)
+    ->  Level = silent
+    ;   Level = informational
     ).
 
 locationprefix(File:Line) -->
     !,
-    [ url(File:Line), ':\n\t' ].
+    [ url(File:Line), ':'-[], nl, '    ' ].
 locationprefix(test(Unit,_Test,Line)) -->
     !,
     { unit_file(Unit, File) },
@@ -1817,12 +1815,14 @@ message(error(plunit(incompatible_options, Tests), _)) -->
     [ 'PL-Unit: incompatible test-options: ~p'-[Tests] ].
 message(plunit(sto(true))) -->
     [ 'Option sto(true) is ignored.  See `occurs_check` option.'-[] ].
+message(plunit(test_files(Total, Loaded))) -->
+    [ 'Found ~D .plt test files, loaded ~D'-[Total, Loaded] ].
 
 					% Unit start/end
 message(plunit(jobs(1))) -->
     !.
 message(plunit(jobs(N))) -->
-    [ 'Tesing with ~D parallel jobs'-[N] ].
+    [ 'Testing with ~D parallel jobs'-[N] ].
 message(plunit(begin(_Unit))) -->
     { tty_feedback },
     !.
@@ -1853,32 +1853,58 @@ message(plunit(fixme([]))) --> !.
 message(plunit(fixme(Tuples))) -->
     !,
     fixme_message(Tuples).
+message(plunit(total_time(Time))) -->
+    [ 'Test run completed'-[] ],
+    test_time(Time).
 
 					% Blocked tests
-message(plunit(blocked(1))) -->
+message(plunit(blocked(1, Tests))) -->
     !,
-    [ 'one test is blocked:'-[] ].
-message(plunit(blocked(N))) -->
-    [ '~D tests are blocked:'-[N] ].
-message(plunit(blocked(Pos, Name, Reason))) -->
+    [ 'one test is blocked'-[] ],
+    blocked_tests(Tests).
+message(plunit(blocked(N, Tests))) -->
+    [ '~D tests are blocked'-[N] ],
+    blocked_tests(Tests).
+
+blocked_tests(Tests) -->
+    { current_test_flag(show_blocked, true) },
+    !,
+    [':'-[]],
+    list_blocked(Tests).
+blocked_tests(_) -->
+    [ ' (use run_tests/2 with ', ansi(code, 'show_blocked(true)', []),
+      ' for details)'-[]
+    ].
+
+list_blocked([]) --> !.
+list_blocked([blocked(_Unit:Test, Pos, Reason)|T]) -->
+    [nl],
     locationprefix(Pos),
-    test_name(Name, -),
-    [ ': ~w'-[Reason] ].
+    test_name(Test, -),
+    [ ': ~w'-[Reason] ],
+    list_blocked(T).
 
 					% fail/success
 message(plunit(no_tests)) -->
     !,
     [ 'No tests to run' ].
-message(plunit(all_passed(1, 1))) -->
+message(plunit(all_passed(1, 1, Time))) -->
     !,
-    [ 'test passed' ].
-message(plunit(all_passed(Total, Total))) -->
+    [ 'test passed' ],
+    test_time(Time).
+message(plunit(all_passed(Total, Total, Time))) -->
     !,
-    [ 'All ~D tests passed'-[Total] ].
-message(plunit(all_passed(Total, Count))) -->
+    [ 'All ~D tests passed'-[Total] ],
+    test_time(Time).
+message(plunit(all_passed(Total, Count, Time))) -->
     !,
     { SubTests is Count-Total },
-    [ 'All ~D (+~D sub-tests) tests passed'-[Total, SubTests] ].
+    [ 'All ~D (+~D sub-tests) tests passed'- [Total, SubTests] ],
+    test_time(Time).
+
+test_time(Time) -->
+    [ ' in ~3f seconds (~3f cpu)'-[Time.wall, Time.cpu] ].
+
 message(plunit(passed(Count))) -->
     !,
     [ '~D tests passed'-[Count] ].
@@ -1909,17 +1935,17 @@ message(plunit(fixme(Failed,Passed,Nondet))) -->
 
 message(plunit(begin(Unit:Test, _Location, Progress))) -->
     { tty_columns(SummaryWidth, _Margin),
-      test_name_summary(Test, SummaryWidth, NameS),
+      test_name_summary(Unit:Test, SummaryWidth, NameS),
       progress_string(Progress, ProgressS)
     },
     (   { tty_feedback,
 	  tty_clear_to_eol(CE)
 	}
-    ->  [ at_same_line, '\r[~w] ~w:~w ..~w'-[ProgressS, Unit, NameS,
+    ->  [ at_same_line, '\r[~w] ~w ..~w'-[ProgressS, NameS,
 					     CE], flush ]
     ;   { jobs(_) }
-    ->  [ '[~w] ~w:~w ..'-[ProgressS, Unit, NameS] ]
-    ;   [ '[~w] ~w:~w ..'-[ProgressS, Unit, NameS], flush ]
+    ->  [ '[~w] ~w ..'-[ProgressS, NameS] ]
+    ;   [ '[~w] ~w ..'-[ProgressS, NameS], flush ]
     ).
 message(plunit(end(_UnitTest, _Location, _Progress))) -->
     [].
@@ -1932,12 +1958,12 @@ message(plunit(progress(Unit:Test, Status, Progress, Time))) -->
     { jobs(_),
       !,
       tty_columns(SummaryWidth, Margin),
-      test_name_summary(Test, SummaryWidth, NameS),
+      test_name_summary(Unit:Test, SummaryWidth, NameS),
       progress_string(Progress, ProgressS),
       progress_tag(Status, Tag, _Keep, Style)
     },
-    [ ansi(Style, '[~w] ~w:~w ~`.t ~w (~3f sec)~*|',
-	   [ProgressS, Unit, NameS, Tag, Time.wall, Margin]) ].
+    [ ansi(Style, '[~w] ~w ~`.t ~w (~3f sec)~*|',
+	   [ProgressS, NameS, Tag, Time.wall, Margin]) ].
 message(plunit(progress(_UnitTest, Status, _Progress, Time))) -->
     { tty_columns(_SummaryWidth, Margin),
       progress_tag(Status, Tag, _Keep, Style)
@@ -2027,31 +2053,35 @@ message(plunit(test_output(_, Output))) -->
 :- if(swi).
 message(interrupt(begin)) -->
     { thread_self(Me),
-      running(Unit, Test, Line, STO, Me),
+      running(Unit, Test, Line, Progress, Me),
       !,
-      unit_file(Unit, File)
+      unit_file(Unit, File),
+      restore_output_state
     },
     [ 'Interrupted test '-[] ],
-    running(running(Unit:Test, File:Line, STO, Me)),
+    running(running(Unit:Test, File:Line, Progress, Me)),
     [nl],
     '$messages':prolog_message(interrupt(begin)).
 message(interrupt(begin)) -->
     '$messages':prolog_message(interrupt(begin)).
 :- endif.
 
+message(concurrent) -->
+    [ 'concurrent(true) at the level of units is currently ignored.', nl,
+      'See set_test_options/1 with jobs(Count) for concurrent testing.'
+    ].
+
 test_name(Name, forall(Bindings, _Nth-I)) -->
     !,
-    [ 'test ~w (~d-th forall bindings = ~p)'-[Name, I, Bindings] ].
+    test_name(Name, -),
+    [ ' (~d-th forall bindings = '-[I],
+      ansi(code, '~p', [Bindings]), ')'-[]
+    ].
 test_name(Name, _) -->
     !,
-    [ 'test ~w'-[Name] ].
+    [ 'test ', ansi(code, '~w', [Name]) ].
 
-det(true) -->
-    [ 'deterministic' ].
-det(false) -->
-    [ 'non-deterministic' ].
-
-running(running(Unit:Test, File:Line, Thread)) -->
+running(running(Unit:Test, File:Line, _Progress, Thread)) -->
     thread(Thread),
     [ '~q:~q at '-[Unit, Test], url(File:Line) ].
 running([H|T]) -->
@@ -2194,12 +2224,16 @@ test_name_summary(Term, MaxLen, Summary) :-
     atom_length(Text, Len),
     (   Len =< MaxLen
     ->  Summary = Text
-    ;   Pre is MaxLen - 8,
+    ;   End is MaxLen//2,
+        Pre is MaxLen - End - 2,
         sub_string(Text, 0, Pre, _, PreText),
-        sub_string(Text, _, 5, 0, PostText),
-        format(string(Summary), '~w...~w', [PreText,PostText])
+        sub_string(Text, _, End, 0, PostText),
+        format(string(Summary), '~w..~w', [PreText,PostText])
     ).
 
+summary_string(Unit:Test, String) =>
+    summary_string(Test, String1),
+    atomics_to_string([Unit, String1], :, String).
 summary_string(@(Name,Vars), String) =>
     format(string(String), '~W (using ~W)',
            [ Name, [numbervars(true), quoted(false)],
@@ -2243,18 +2277,32 @@ progress_tag(forall(_,_),   Tag, Keep, Style) =>
     Tag = '**FAILED', Keep = true, Style = error.
 
 
+		 /*******************************
+		 *           OUTPUT		*
+		 *******************************/
+
+save_output_state :-
+    stream_property(Output, alias(user_output)),
+    stream_property(Error, alias(user_error)),
+    asserta(output_streams(Output, Error)).
+
+restore_output_state :-
+    output_streams(Output, Error),
+    !,
+    set_stream(Output, alias(user_output)),
+    set_stream(Error, alias(user_error)).
+restore_output_state.
+
+
+
 		 /*******************************
 		 *      CONCURRENT STATUS       *
 		 *******************************/
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-
-This part deals with interactive feedback when we are running multiple
- threads.   The terminal  window  cannot  work on  top  of the  Prolog
- message  infrastructure and  (thus)  we have  to  use more  low-level
- means.
-
-
+This part deals with interactive feedback   when we are running multiple
+threads. The terminal window cannot work on   top  of the Prolog message
+infrastructure and (thus) we have to use more low-level means.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 :- dynamic
@@ -2285,11 +2333,11 @@ job_feedback(Level, Msg) :-
 
 job_feedback(begin(Unit:Test, _Location, Progress)) =>
     tty_columns(SummaryWidth, _Margin),
-    test_name_summary(Test, SummaryWidth, NameS),
+    test_name_summary(Unit:Test, SummaryWidth, NameS),
     progress_string(Progress, ProgressS),
     tty_clear_to_eol(CE),
-    job_format(comment, '\r[~w] ~w:~w ..~w',
-	       [ProgressS, Unit, NameS, CE]),
+    job_format(comment, '\r[~w] ~w ..~w',
+	       [ProgressS, NameS, CE]),
     flush_output.
 job_feedback(end(_UnitTest, _Location, _Progress)) =>
     true.
@@ -2407,8 +2455,7 @@ job_self(Job) :-
 
 tty_feedback :-
     has_tty,
-    current_test_flag(test_options, Options),
-    option(format(tty), Options, tty).
+    current_test_flag(format, tty).
 
 has_tty :-
     stream_property(user_output, tty(true)).
@@ -2416,11 +2463,12 @@ has_tty :-
 tty_columns(SummaryWidth, Margin) :-
     tty_width(W),
     Margin is W-8,
-    SummaryWidth = max(20,Margin-50).
+    SummaryWidth is max(20,Margin-34).
 
 tty_width(W) :-
     current_predicate(tty_size/2),
     catch(tty_size(_Rows, Cols), error(_,_), fail),
+    Cols > 25,
     !,
     W = Cols.
 tty_width(80).
@@ -2428,9 +2476,11 @@ tty_width(80).
 tty_header_line(Width) :-
     ansi_format(comment, '~N~`\u2015t~*|~n', [Width]).
 
+:- if(current_predicate(tty_get_capability/3)).
 tty_clear_to_eol(S) :-
     tty_get_capability(ce, string, S),
     !.
+:- endif.
 tty_clear_to_eol('\e[K').
 
 tty_up_and_clear(Lines) :-
