@@ -78,6 +78,22 @@ foo(Baz) :- baz.
     (should (not (sweeprolog-beginning-of-next-top-term)))
     (should (= (point) 19))))
 
+(ert-deftest help-echo-for-dependency ()
+  "Test that the `help-echo' property is set correctly."
+  (let ((temp (make-temp-file "sweeprolog-help-echo-text"
+                              nil
+                              "pl"
+                              "
+:- use_module(library(lists)).
+
+foo(Foo, Bar) :- flatten(Bar, Baz), member(Foo, Baz).
+")))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (goto-char 24)
+    (should (string-match "Dependency on .*, resolves calls to flatten/2, member/2"
+                          (help-at-pt-kbd-string)))))
+
 (ert-deftest terms-at-point ()
   "Test `sweeprolog-term-search'."
   (let ((temp (make-temp-file "sweeprolog-terms-at-point-test"
@@ -108,6 +124,10 @@ recursive(Var) :-
     *-> Bar is foo
     )")))))
 
+(ert-deftest predicate-location ()
+  "Test `sweeprolog-predicate-location'."
+  (should (sweeprolog-predicate-location "memory_file:new_memory_file/1")))
+
 (ert-deftest term-search ()
   "Test `sweeprolog-term-search'."
   (let ((temp (make-temp-file "sweeprolog-test"
@@ -216,7 +236,6 @@ foo(Foo) :- bar.
     (should (equal (get-text-property (+ (point-min) 16)
                                       'font-lock-face)
                    '(sweeprolog-local-default-face
-                     sweeprolog-predicate-indicator-default-face
                      sweeprolog-body-default-face)))
     (should (equal (get-text-property (+ (point-min) 23)
                                       'font-lock-face)
@@ -234,6 +253,20 @@ foo(Foo) :- bar.
                    '(sweeprolog-undefined-default-face
                      sweeprolog-body-default-face)))))
 
+(ert-deftest yank-hole ()
+  "Test killing and yanking a hole as a plain variable."
+  (let ((temp (make-temp-file "sweeprolog-test"
+                              nil
+                              "pl"
+                              "")))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (sweeprolog-insert-term-with-holes ":-" 2)
+    (should (get-text-property (point-min) 'sweeprolog-hole))
+    (call-interactively #'kill-region)
+    (call-interactively #'yank)
+    (should (not (get-text-property (point-min) 'sweeprolog-hole)))))
+
 (ert-deftest insert-term-with-holes ()
   "Test `sweeprolog-insert-term-with-holes'."
   (let ((temp (make-temp-file "sweeprolog-test"
@@ -254,6 +287,89 @@ foo(Foo) :- bar.
     (should (string= (buffer-string)
                      "foo(bar, (_->_;_), _):-_."))))
 
+(ert-deftest rename-variable ()
+  "Tests renaming varialbes."
+  (let ((temp (make-temp-file "sweeprolog-test"
+                              nil
+                              ".pl"
+                              "foo(Bar,Baz) :- spam(Baz,Bar).")))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (goto-char (point-min))
+    (sweeprolog-rename-variable "Bar" "Spam")
+    (sweeprolog-rename-variable "Baz" "Bar")
+    (sweeprolog-rename-variable "Spam" "Baz")
+    (should (string= (buffer-string)
+                     "foo(Baz,Bar) :- spam(Bar,Baz)."))))
+
+(ert-deftest increment-variable ()
+  "Tests renaming varialbes."
+  (let ((temp (make-temp-file "sweeprolog-test"
+                              nil
+                              ".pl"
+                              "
+foo(Bar0,Bar) :-
+    spam(Bar0,Bar1),
+    bar(Bar1,Bar2),
+    baz(Bar2, Bar).
+")))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (goto-char (1+ (point-min)))
+    (sweeprolog-increment-numbered-variables 1 (point) "Bar1")
+    (should (string= (buffer-string)
+                     "
+foo(Bar0,Bar) :-
+    spam(Bar0,Bar2),
+    bar(Bar2,Bar3),
+    baz(Bar3, Bar).
+"))))
+
+(ert-deftest find-references ()
+  "Tests `sweeprolog-predicate-references'."
+  (let ((temp (make-temp-file "sweeprolog-test"
+                              nil
+                              ".pl"
+                              ":- module(test_sweep_find_references, [caller/0]).
+
+caller :- callee, baz, callee.
+caller :- baz, callee, baz.
+
+callee.
+
+baz.
+"
+                              )))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (should (equal (sweeprolog-predicate-references "test_sweep_find_references:callee/0")
+                   (list (list "caller/0" temp 63 6)
+                         (list "caller/0" temp 76 6)
+                         (list "caller/0" temp 99 6))))))
+
+(ert-deftest forward-many-holes ()
+  "Tests jumping over holes with `sweeprolog-forward-hole'."
+  (let ((temp (make-temp-file "sweeprolog-test"
+                              nil
+                              ".pl"
+                              "\n"
+                              )))
+    (find-file-literally temp)
+    (sweeprolog-mode)
+    (goto-char (point-min))
+    (sweeprolog-insert-term-with-holes ":-" 2)
+    (deactivate-mark)
+    (goto-char (point-max))
+    (sweeprolog-insert-term-with-holes ":-" 2)
+    (goto-char (point-min))
+    (should (= (sweeprolog-count-holes) 4))
+    (sweeprolog-forward-hole 2)
+    (should (= (point) 5))
+    (sweeprolog-forward-hole -1)
+    (should (= (point) 2))
+    (sweeprolog-forward-hole -2)
+    (should (= (point) 8))))
+
 (ert-deftest plunit-testset-skeleton ()
   "Tests inserting PlUnit test-set blocks."
   (let ((temp (make-temp-file "sweeprolog-test"
@@ -387,7 +503,7 @@ baz(Baz) :- findall(X, b_g
     (call-interactively #'completion-at-point)
     (should (string= (buffer-string)
                      "
-baz(Baz) :- findall(X, b_getval(_, _)
+baz(Baz) :- findall(X, b_getval(Name, Value)
 "
                      ))))
 
