@@ -73,8 +73,8 @@
 #define PUBLIC_GLOBAL extern		/* exported global variables */
 #endif
 
-#if defined(WIN32) || defined(__CYGWIN__)
-#define __pce_export __declspec(dllexport)
+#if defined(WIN32)
+#define __pce_export __declspec(dllexport) extern
 #else
 #define __pce_export extern
 #endif
@@ -257,6 +257,10 @@ cpdata(to, from, type, n)
 #include O_EXTRA_SYSTEM_TYPES
 #endif
 
+#define functionToPtr(f)   ((void*)(intptr_t)(f))
+#define PtrToFunction(ptr) ((Any(*)())(intptr_t)(ptr))
+
+
 		 /*******************************
 		 * OS-IDENTIFIERS (STRICT_ANSI) *
 		 *******************************/
@@ -318,33 +322,14 @@ typedef struct dCell	      **DelegateList;   /* See msg-passing.c */
 
 #define INVOKE_FUNC ((SendFunc)~0L)
 
-		/********************************
-		*	    POINTERS		*
-		********************************/
-
-#ifdef VARIABLE_POINTER_OFFSET
-#undef POINTER_OFFSET
-GLOBAL uintptr_t pce_data_pointer_offset;
-#define POINTER_OFFSET pce_data_pointer_offset
-#else
-#ifndef POINTER_OFFSET
-#define POINTER_OFFSET (0L)
-#endif
-#endif
-
-#define PointerToCInt(p) (((uintptr_t)(p) - POINTER_OFFSET)/sizeof(void*))
-#define PointerToInt(p)	 toInt(PointerToCInt(p))
-#define longToPointer(i) ((Any) (i * sizeof(void*) + POINTER_OFFSET))
-#define IntToPointer(i)  longToPointer(valInt(i))
-
 
 		/********************************
 		*           TAG MASKS		*
 		********************************/
 
-#define INT_MASK	0x00000001	/* 10 mask for Int (integers) */
-#define MASK_MASK	0x00000001	/* 11 Mask Mask */
-#define TAG_BITS	1		/* number of mask bits for INT */
+#define INT_MASK	((uintptr_t)0x01) /* 01 mask for Int (integers) */
+#define MASK_MASK	((uintptr_t)0x01) /* 11 Mask Mask */
+#define TAG_BITS	1		  /* number of mask bits for INT */
 
 #define MaskOf(obj)		((uintptr_t)(obj) & MASK_MASK)
 #define UnMask(obj)		((uintptr_t)(obj) & ~MASK_MASK)
@@ -384,23 +369,6 @@ integer is declared as of type Int (for casting purposes). The following
 test, conversion and computation macro's are provided.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-/* TBD: INTPTR_FORMAT is available as PRTdPTR from <inttypes.h>.  Our
-   include for fixed size integers is a bit flaky on various platforms
-   though
-*/
-
-#ifndef INTPTR_FORMAT		/* printf format for intptr_t */
-#if SIZEOF_INT == SIZEOF_VOIDP
-#define INTPTR_FORMAT "%d"
-#elif SIZEOF_LONG == SIZEOF_VOIDP
-#define INTPTR_FORMAT "%ld"
-#elif defined(__WIN64)
-#define INTPTR_FORMAT "%I64d"
-#else
-#error "Config needs INTPTR_FORMAT"
-#endif
-#endif
-
 #undef max
 #undef min
 #define max(a, b)	((a) > (b) ? (a) : (b))
@@ -922,7 +890,7 @@ typedef struct _classdecl
 #define GM(n, a, r, t, f, g, s) { n, a, r, t, (GetFunc) f, (Name) g, s }
 #define RC(n, t, d, s)		{ n, t, d, s }
 #define IV(n, t, f, g, s)	{ n, t, f, NULL,        (Name) g, s }
-#define SV(n, t, f, c, g, s)	{ n, t, f, (void *)(c), (Name) g, s }
+#define SV(n, t, f, c, g, s)	{ n, t, f, functionToPtr(c), (Name) g, s }
 #define IVEntries(l)		(sizeof(l) / sizeof(vardecl))
 #define SMEntries(l)		(sizeof(l) / sizeof(senddecl))
 #define GMEntries(l)		(sizeof(l) / sizeof(getdecl))
@@ -1024,8 +992,8 @@ NewClass(type)
 End;
 
 NewClass(constraint)
-  Any	        from;			/* 'From' object of constraint */
-  Any	        to;			/* 'To' object of constraint */
+  Any		from;			/* 'From' object of constraint */
+  Any		to;			/* 'To' object of constraint */
   Relation	relation;		/* relation they have */
   Name		locked;			/* locked fro further messages? */
 End;
@@ -1519,7 +1487,7 @@ struct var_extension
     } else \
     { _var_env.size = 0; \
       for(_i=0; _i<ac; _i++) \
-        assignVar(Arg(_i+1), (av)[_i], DEFAULT); \
+	assignVar(Arg(_i+1), (av)[_i], DEFAULT); \
     } \
  \
     code; \
@@ -1570,6 +1538,32 @@ typedef intptr_t AnswerMark;
 #include "../adt/proto.h"
 #include "../rel/proto.h"
 
+/* Pointer <-> integer conversion */
+
+#ifdef VARIABLE_POINTER_OFFSET
+#undef POINTER_OFFSET
+GLOBAL uintptr_t pce_data_pointer_offset;
+#define POINTER_OFFSET pce_data_pointer_offset
+#else
+#ifndef POINTER_OFFSET
+#define POINTER_OFFSET (0L)
+#endif
+#endif
+
+#define longToPointer(i) ((Any) (i * sizeof(void*) + POINTER_OFFSET))
+#define IntToPointer(i)  longToPointer(valInt(i))
+
+static inline uintptr_t
+PointerToCInt(void *p)
+{ uintptr_t i = ((uintptr_t)(p) - POINTER_OFFSET)/sizeof(void*);
+
+  assert(longToPointer(i) == p);
+  return i;
+}
+
+#define PointerToInt(p)	 toInt(PointerToCInt(p))
+
+
 #define getSubName(n, f, t) (Name)getSubCharArray((CharArray)(n), f, t)
 
 					/* Interface callback stubs */
