@@ -3,9 +3,10 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2020, University of Amsterdam
+    Copyright (c)  1985-2022, University of Amsterdam
                               VU University Amsterdam
 			      CWI, Amsterdam
+			      SWI-Prolog Solutions b.v
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -119,7 +120,7 @@ static const PL_extension foreigns[] = {
   FRG("write_term",		3, pl_write_term3,	 META|ISO),
   FRG("write",			1, pl_write,		      ISO),
   FRG("writeq",			1, pl_writeq,		      ISO),
-  FRG("writeln",		1, pl_writeln,		        0),
+  FRG("writeln",		1, pl_writeln,			0),
   FRG("print",			1, pl_print,			0),
 
   FRG("read",			1, pl_read,		      ISO),
@@ -165,7 +166,7 @@ static const PL_extension foreigns[] = {
   FRG("$dwim_predicate",	2, pl_dwim_predicate,	     NDET),
 
 #ifdef O_PROLOG_HOOK
-  FRG("set_prolog_hook",	3, pl_set_prolog_hook,	        0),
+  FRG("set_prolog_hook",	3, pl_set_prolog_hook,		0),
 #endif
   FRG("context_module",		1, pl_context_module,	     META),
 
@@ -189,7 +190,7 @@ static const PL_extension foreigns[] = {
 
   FRG("read",			2, pl_read2,		      ISO),
   FRG("write",			2, pl_write2,		      ISO),
-  FRG("writeln",		2, pl_writeln2,		        0),
+  FRG("writeln",		2, pl_writeln2,			0),
   FRG("writeq",			2, pl_writeq2,		      ISO),
   FRG("print",			2, pl_print2,			0),
   FRG("write_canonical",	2, pl_write_canonical2,	      ISO),
@@ -197,7 +198,7 @@ static const PL_extension foreigns[] = {
 
   FRG("format_predicate",	2, pl_format_predicate,	     META),
   FRG("current_format_predicate", 2, pl_current_format_predicate,
-						        META|NDET),
+							META|NDET),
   FRG("get_time",		1, pl_get_time,			0),
 
 #ifdef O_PLMT
@@ -221,9 +222,9 @@ static unsigned int
 predicate_signature(const Definition def)
 { char str[256];
 
-  Ssprintf(str, "%s/%d/%d",
+  Ssprintf(str, "%s/%zd/0x%" PRIx64,
 	   stringAtom(def->functor->name),
-	   (int)def->functor->arity,
+	   def->functor->arity,
 	   def->flags);
 
   return MurmurHashAligned2(str, strlen(str), SIGNATURE_SEED);
@@ -390,6 +391,7 @@ DECL_PLIST(wic);
 DECL_PLIST(attvar);
 DECL_PLIST(gvar);
 DECL_PLIST(win);
+DECL_PLIST(apple);
 DECL_PLIST(file);
 DECL_PLIST(files);
 DECL_PLIST(glob);
@@ -422,13 +424,17 @@ DECL_PLIST(event);
 DECL_PLIST(transaction);
 DECL_PLIST(undo);
 DECL_PLIST(error);
+DECL_PLIST(coverage);
+#ifdef __EMSCRIPTEN__
+DECL_PLIST(wasm);
+#endif
 
 void
 initBuildIns(void)
 { ExtensionCell ecell;
   Module m = MODULE_system;
 
-  GD->procedures.dirty = newHTable(32);
+  initProcedures();
 
   registerBuiltins(foreigns);
   REG_PLIST(alloc);
@@ -479,6 +485,9 @@ initBuildIns(void)
   REG_PLIST(win);
   REG_PLIST(dde);
 #endif
+#ifdef __APPLE__
+  REG_PLIST(apple);
+#endif
 #ifdef O_LOCALE
   REG_PLIST(locale);
 #endif
@@ -495,6 +504,12 @@ initBuildIns(void)
   REG_PLIST(transaction);
   REG_PLIST(undo);
   REG_PLIST(error);
+#ifdef O_COVERAGE
+  REG_PLIST(coverage);
+#endif
+#ifdef __EMSCRIPTEN__
+  REG_PLIST(wasm);
+#endif
 
 #define LOOKUPPROC(name) \
 	{ GD->procedures.name = lookupProcedure(FUNCTOR_ ## name, m); \
@@ -521,6 +536,8 @@ initBuildIns(void)
 #ifdef O_ATTVAR
   LOOKUPPROC(dwakeup1);
 #endif
+  GD->procedures.heartbeat0 = lookupProcedure(FUNCTOR_heartbeat0,
+					      PL_new_module(PL_new_atom("prolog")));
   PROCEDURE_exception_hook4  =
 	PL_predicate("prolog_exception_hook", 4, "user");
   PROCEDURE_tune_gc3 =
